
from abuseACL.structures.ADObject import ADObject

class ADgMSA(ADObject):

    def __init__(self, distinguishedName: str, sAMAccountName: str, objectSid: str, nTSecurityDescriptor: bytes, memberOf: list[bytes]) -> None:
        self.distinguishedName      = distinguishedName[0].decode()
        self.sAMAccountName         = sAMAccountName[0].decode().lower()
        self.objectSid              = self.convertSid(objectSid[0])
        self.nTSecurityDescriptor   = self.parseSecurityDescriptor(nTSecurityDescriptor[0])
        self.memberOf               = [group.decode() for group in memberOf]

    @staticmethod
    def getgMSASid(gMSAs: list, principal: str) -> str:
        for gMSA in gMSAs:
            gMSA: ADgMSA

            if gMSA.sAMAccountName == principal:
                return gMSA.objectSid

        return None

    @staticmethod
    def getgMSAFromSid(gMSAs: list[ADObject], gMSASid: str):
        for gMSA in gMSAs:
            gMSA: ADgMSA

            if gMSA.objectSid == gMSASid:
                return gMSA




from ldap3.protocol.formatters.formatters import format_sid
from impacket.ldap.ldaptypes import SR_SECURITY_DESCRIPTOR

class ADObject:

    def __init__(self) -> None:
        pass

    def parseSecurityDescriptor(self, nTSecurityDescriptor: bytes) -> SR_SECURITY_DESCRIPTOR:
        return SR_SECURITY_DESCRIPTOR(nTSecurityDescriptor)

    def convertSid(self, objectSid: bytes) -> str:
        return format_sid(objectSid)

    def __str__(self) -> str:
        return "{0: <30} {1}".format(self.sAMAccountName, self.objectSid)




class Credentials:

    def __init__(self, username: str, password: str, domain: str, ntlmhash: str, aesKey: str, doKerberos: bool, doSimpleBind: bool) -> None:
        self.username = username
        self.password = password
        self.domain = domain
        self.ntlmhash = ntlmhash
        self.aesKey = aesKey
        self.doKerberos = doKerberos
        self.doSimpleBind = doSimpleBind

    def getAuthenticationSecret(self) -> str:
        return self.password or self.ntlmhash




class Target:

    tlsv1_2: bool = None
    tlsv1: bool = None

    def __init__(self, remote: str, port: int) -> None:
        self.remote = remote
        self.port = port

    def use_tls(self) -> bool:
        return self.tlsv1_2 or self.tlsv1




from abuseACL.structures.ADObject import ADObject

class ADGroup(ADObject):

    def __init__(self, distinguishedName: list[bytes], name: list[bytes], sAMAccountName: list[bytes], objectSid: list[bytes], nTSecurityDescriptor: list[bytes], memberOf: list[bytes]) -> None:
        self.distinguishedName      = distinguishedName[0].decode()
        self.name                   = name[0].decode()
        self.sAMAccountName         = sAMAccountName[0].decode().lower()
        self.objectSid              = self.convertSid(objectSid[0]) if isinstance(objectSid, list) else objectSid
        self.nTSecurityDescriptor   = self.parseSecurityDescriptor(nTSecurityDescriptor[0]) if nTSecurityDescriptor else None
        self.memberOf               = [group.decode() for group in memberOf]

    @staticmethod
    def getGroupSid(groups: list, groupname: str) -> str:
        for group in groups:
            group: ADGroup

            if group.sAMAccountName == groupname:
                return group.objectSid

        return None

    @staticmethod
    def getGroupFromSid(groups: list[ADObject], groupSid: str):
        for group in groups:
            group: ADGroup
        
            if group.objectSid == groupSid:
                return group

    @staticmethod
    def getGroupSidFromDistinguishedName(groups: list[ADObject], groupDN: str) -> ADObject:
        for group in groups:
            group: ADGroup

            if group.distinguishedName.lower() == groupDN.lower():
                return group




from colorama import Style, Fore

import time

class Logger:

    def __init__(self, debug: bool, timestamp: bool) -> None:
        self.__debug = debug
        self.__timestamp = timestamp

    def __toStdout(self, color: str, title: str, msg: str) -> None:
        timestamp = str()

        if self.__timestamp:
            timestamp = time.strftime("[%Y/%m/%d %H:%M:%S] ")

        print("%s%s[%s] %s%s" % (color, timestamp, title, msg, Style.RESET_ALL))

    def debug(self, msg: str) -> None:
        if self.__debug:
            self.__toStdout(Fore.BLUE, "d", msg)

    def error(self, msg: str) -> None:
        self.__toStdout(Fore.RED, "!", msg)

    def vuln(self, msg: str) -> None:
        self.__toStdout(Fore.GREEN, "*", msg)

    def print(self, msg: str) -> None:
        self.__toStdout("", "i", msg)




from abuseACL.structures.ADObject import ADObject

class ADDomain(ADObject):

    def __init__(self, distinguishedName: list[bytes], objectSid: list[bytes], nTSecurityDescriptor: list[bytes]) -> None:
        self.distinguishedName      = distinguishedName[0].decode()
        self.objectSid              = self.convertSid(objectSid[0])
        self.nTSecurityDescriptor   = self.parseSecurityDescriptor(nTSecurityDescriptor[0])
        self.sAMAccountName         = self.distinguishedName.replace("DC=", "").replace(",", ".")




from abuseACL.structures.ADObject import ADObject

class ADComputer(ADObject):

    def __init__(self, distinguishedName: str, name: str, sAMAccountName: str, objectSid: str, nTSecurityDescriptor: bytes, userAccountControl: int, memberOf: list[bytes]) -> None:
        self.distinguishedName      = distinguishedName[0].decode()
        self.name                   = name[0].decode()
        self.sAMAccountName         = sAMAccountName[0].decode().lower()
        self.objectSid              = self.convertSid(objectSid[0])
        self.nTSecurityDescriptor   = self.parseSecurityDescriptor(nTSecurityDescriptor[0])
        self.userAccountControl     = int(userAccountControl[0].decode())
        self.memberOf               = [group.decode() for group in memberOf]

        self.isUserEnable = self.userAccountControl & 0x0002

    @staticmethod
    def getComputerSid(computers: list, computername: str) -> str:
        for computer in computers:
            computer: ADComputer

            if computer.sAMAccountName == computername:
                return computer.objectSid

        return None

    @staticmethod
    def getComputerFromSid(computers: list[ADObject], computerSid: str):
        for computer in computers:
            computer: ADComputer

            if computer.objectSid == computerSid:
                return computer




##
# https://learn.microsoft.com/fr-fr/windows/win32/adschema/a-samaccounttype
##

class sAMAccountType:

    SAM_DOMAIN_OBJECT = 0x0
    SAM_GROUP_OBJECT = 0x10000000
    SAM_NON_SECURITY_GROUP_OBJECT = 0x10000001
    SAM_ALIAS_OBJECT = 0x20000000
    SAM_NON_SECURITY_ALIAS_OBJECT = 0x20000001
    SAM_USER_OBJECT = 0x30000000
    SAM_NORMAL_USER_ACCOUNT = 0x30000000
    SAM_MACHINE_ACCOUNT = 0x30000001
    SAM_TRUST_ACCOUNT = 0x30000002
    SAM_APP_BASIC_GROUP = 0x40000000
    SAM_APP_QUERY_GROUP = 0x40000001
    SAM_ACCOUNT_TYPE_MAX = 0x7fffffff




from typing import List

import ssl as tls
import ldap3

from abuseACL.structures.sAMAccountType import sAMAccountType
from abuseACL.structures.Credentials import Credentials
from abuseACL.structures.Target import Target
from abuseACL.structures.ADObject.ADCertificateTemplate import ADCertificateTemplate
from abuseACL.structures.ADObject.ADAdminSDHolder import ADAdminSDHolder
from abuseACL.structures.ADObject.ADContainer import ADContainer
from abuseACL.structures.ADObject.ADComputer import ADComputer
from abuseACL.structures.ADObject.ADDomain import ADDomain
from abuseACL.structures.ADObject.ADSchema import ADSchema
from abuseACL.structures.ADObject.ADGroup import ADGroup
from abuseACL.structures.ADObject.ADUser import ADUser
from abuseACL.structures.ADObject.ADgMSA import ADgMSA
from abuseACL.structures.ADObject.ADGPO import ADGPO
from abuseACL.structures.ADObject.ADOU import ADOU
from abuseACL.structures.ADObject import ADObject
from abuseACL.structures.structures import WELL_KNOWN_SID
from abuseACL.network.Kerberos import Kerberos
from abuseACL.core.Logger import Logger

class LDAP:

    users = list()
    groups = list()
    computers = list()
    certificatesTemplates = list()
    gpos = list()
    ous = list()
    containers = list()
    adminSDHolder = list()
    schema = list()
    gMSA = list()
    domain: ADDomain = None

    def __init__(self, forest: str, target: Target, credentials: Credentials, logger: Logger) -> None:
        self.target         = target
        self.credentials    = credentials
        self.logger         = logger

        self.__getPort()
        self.__checkAuthentication()

    def __getPort(self) -> None:
        if self.target.port:
            return

        self.target.port, self.target.tlsv1_2 = self.__tryLDAPS(tls.PROTOCOL_TLSv1_2, self.target.port)

        if self.target.tlsv1_2 is None:
            self.target.port, self.target.tlsv1 = self.__tryLDAPS(tls.PROTOCOL_TLSv1, self.target.port)

            if self.target.tlsv1 is None:
                self.target.port = self.__tryLDAP(self.target.port)

        if self.target.port is None:
            self.logger.error(f"Impossible to communicate with the target {self.target.remote} !")
            exit(1)

    def __checkAuthentication(self) -> None:
        self.logger.print("Trying to connect to %s:%d" % (self.target.remote, self.target.port))
        self.__Authentication()

        try:
            self.getNamingContexts()
        except IndexError:
            self.logger.error("Invalid credentials !")
            exit(1)

        self.logger.print("Authentication success !")

    def __Authentication(self) -> ldap3.Connection:
        user = "%s\\%s" % (self.credentials.domain, self.credentials.username)

        ldapTls = None

        if self.target.tlsv1_2:
            ldapTls = ldap3.Tls(validate=tls.CERT_NONE, version=tls.PROTOCOL_TLSv1_2, ciphers='ALL:@SECLEVEL=0')
        elif self.target.tlsv1:
            ldapTls = ldap3.Tls(validate=tls.CERT_NONE, version=tls.PROTOCOL_TLSv1, ciphers='ALL:@SECLEVEL=0')
        
        ldapServer = ldap3.Server(self.target.remote, use_ssl=self.target.use_tls(), port=self.target.port, get_info=ldap3.ALL, tls=ldapTls)

        if self.credentials.doKerberos:
            ldapConn = ldap3.Connection(ldapServer)
            ldapConn = self.kerberosAuthentication(ldapConn)
        elif self.credentials.doSimpleBind:
            ldapConn = ldap3.Connection(ldapServer, user=user, password=self.credentials.getAuthenticationSecret(), authentication=ldap3.SIMPLE)
            ldapConn.bind()
        else:
            ldapConn = ldap3.Connection(ldapServer, user=user, password=self.credentials.getAuthenticationSecret(), authentication=ldap3.NTLM)
            try:
                ldapConn.bind()
            except ldap3.core.exceptions.LDAPSocketReceiveError:
                self.logger.error("Can't connect using NTLM, try with -simple-bind option")
                exit(1)

        if ldapConn.result["description"] == "invalidCredentials":
            self.logger.error("Invalid credentials !")
            exit(1)

        return ldapConn

    def __tryLDAPS(self, proto: tls._SSLMethod, port: int) -> int:
        port = port or 636

        ldapTls = ldap3.Tls(validate=tls.CERT_NONE, version=proto, ciphers="ALL:@SECLEVEL=0")
        ldapServer = ldap3.Server(self.target.remote, use_ssl=True, port=port, get_info=ldap3.ALL, tls=ldapTls)
        ldapConn = ldap3.Connection(ldapServer)

        try:
            ldapConn.bind()
        except ldap3.core.exceptions.LDAPSocketOpenError:
            return None, None
        except ldap3.core.exceptions.LDAPSocketReceiveError:
            pass

        return port, True

    def __tryLDAP(self, port: int) -> int:
        self.logger.debug("LDAPS failed, trying with LDAP.")
        port = port or 389

        ldapServer = ldap3.Server(self.target.remote, use_ssl=False, port=port, get_info=ldap3.ALL)
        ldapConn = ldap3.Connection(ldapServer)

        try:
            ldapConn.bind()
        except ldap3.core.exceptions.LDAPSocketOpenError:
            return None
        except ldap3.core.exceptions.LDAPSocketReceiveError:
            return port

        return port

    def kerberosAuthentication(self, ldapConn: ldap3.Connection) -> None:
        blob = Kerberos.kerberosLogin(self.target.remote, self.credentials.username, self.credentials.password,
                                    self.credentials.domain, self.credentials.ntlmhash, self.credentials.aesKey,
                                    kdcHost=self.target.remote)

        request = ldap3.operation.bind.bind_operation(ldapConn.version, ldap3.SASL, self.credentials.username, None, "GSS-SPNEGO", blob.getData())

        # Done with the Kerberos saga, now let's get into LDAP
        # try to open connection if closed
        if ldapConn.closed:
            ldapConn.open(read_server_info=False)

        ldapConn.sasl_in_progress = True
        response = ldapConn.post_send_single_response(ldapConn.send('bindRequest', request, None))

        ldapConn.sasl_in_progress = False

        if response[0]['result'] != 0:
            raise Exception(response)

        ldapConn.bound = True

        return ldapConn
    
    def search(self, dn: str, filter: str, scope: str, attributes: list = ["*"], cookie: str | bytes | None = None) -> list:
        entries = list()

        ldapConn = self.__Authentication()
        ldapConn.search(
            search_base=dn,
            search_filter=filter,
            search_scope=scope,
            attributes=attributes,
            # Controls to get nTSecurityDescriptor from standard user
            # OWNER_SECURITY_INFORMATION + GROUP_SECURITY_INFORMATION + DACL_SECURITY_INFORMATION
            controls=[("1.2.840.113556.1.4.801", True, "%c%c%c%c%c" % (48, 3, 2, 1, 7), )],
            paged_size=5000,
            paged_cookie=cookie
        )
        entries.extend(ldapConn.response)

        if ldapConn.result.get("controls", False):
            cookie = ldapConn.result['controls']['1.2.840.113556.1.4.319']['value']['cookie']
        else:
            cookie = None

        if cookie:
            entries.extend(self.search(dn, filter, scope, attributes, cookie))

        return entries

    def __createArrayOfObject(self, response: list, obj: object) -> list:

        array = list()

        for entry in response:
            # Not a response object
            if entry["type"] != "searchResEntry":
                continue

            array.append(
                obj(**entry["raw_attributes"])
            )

        return array

    def getNamingContexts(self) -> list:
        response = self.search(
            "",
            "(objectClass=*)",
            ldap3.BASE,
            ["namingContexts"]
        )

        self.namingContexts = response[0]["attributes"]["namingContexts"]
        self.defaultNamingContext = self.namingContexts[0]
        self.configurationNamingContext = self.namingContexts[1]
        self.schemaNamingContext = self.namingContexts[2]
        self.domainDnsZonesNamingContext = self.namingContexts[3]
        self.forestDnsZonesNamingContext = self.namingContexts[4]

    def getAllUsers(self) -> List[ADUser]:
        if len(self.users):
            return self.users
        self.logger.print("Retrive all users")

        response = self.search(
            self.defaultNamingContext,
            "(sAMAccountType=%d)" % (sAMAccountType.SAM_NORMAL_USER_ACCOUNT),
            ldap3.SUBTREE,
            ["DistinguishedName", "name", "userPrincipalName", "sAMAccountName", "objectSid", "ntSecurityDescriptor", "userAccountControl", "memberOf"]
        )

        self.users = self.__createArrayOfObject(response, ADUser)

        return self.users

    def getAllGroups(self) -> List[ADGroup]:
        if len(self.groups):
            return self.groups
        self.logger.print("Retrive all groups")

        response = self.search(
            self.defaultNamingContext,
            "(|(sAMAccountType=%d)(sAMAccountType=%d)(sAMAccountType=%d)(sAMAccountType=%d))" % (
                sAMAccountType.SAM_GROUP_OBJECT,
                sAMAccountType.SAM_NON_SECURITY_GROUP_OBJECT,
                sAMAccountType.SAM_ALIAS_OBJECT,
                sAMAccountType.SAM_NON_SECURITY_ALIAS_OBJECT
            ),
            ldap3.SUBTREE,
            ["DistinguishedName", "name", "sAMAccountName", "objectSid", "ntSecurityDescriptor", "memberOf"]
        )

        self.groups = self.__createArrayOfObject(response, ADGroup)
        
        return self.groups
    
    def getAllComputers(self) -> List[ADComputer]:
        if len(self.computers):
            return self.computers
        self.logger.print("Retrive all computers")

        response = self.search(
            self.defaultNamingContext,
            "(sAMAccountType=%d)" % (sAMAccountType.SAM_MACHINE_ACCOUNT),
            ldap3.SUBTREE,
            ["DistinguishedName", "name", "sAMAccountName", "objectSid", "ntSecurityDescriptor", "userAccountControl", "memberOf"]
        )

        self.computers = self.__createArrayOfObject(response, ADComputer)

        return self.computers
    
    def getAllCertificatesTemplates(self) -> List[ADCertificateTemplate]:
        if len(self.certificatesTemplates):
            return self.certificatesTemplates
        self.logger.print("Retrive all certificateTemplates")

        response = self.search(
            f"CN=Certificate Templates,CN=Public Key Services,CN=Services,{self.configurationNamingContext}",
            "(objectClass=pkiCertificateTemplate)",
            ldap3.SUBTREE,
            ["DistinguishedName", "name", "ntSecurityDescriptor"]
        )

        self.certificatesTemplates = self.__createArrayOfObject(response, ADCertificateTemplate)

        return self.certificatesTemplates
    
    def getAllGPOs(self) -> List[ADGPO]:
        if len(self.gpos):
            return self.gpos
        self.logger.print("Retrive all GPOs")

        response = self.search(
            f"CN=Policies,CN=System,{self.defaultNamingContext}",
            "(objectClass=groupPolicyContainer)",
            ldap3.SUBTREE,
            ["DistinguishedName", "displayName", "gPCFileSysPath", "ntSecurityDescriptor"]
        )

        self.gpos = self.__createArrayOfObject(response, ADGPO)

        return self.gpos

    def getAllOUs(self) -> List[ADOU]:
        if len(self.ous):
            return self.ous
        self.logger.print("Retrive all OUs")

        response = self.search(
            self.defaultNamingContext,
            "(objectClass=organizationalUnit)",
            ldap3.SUBTREE,
            ["DistinguishedName", "name", "ntSecurityDescriptor"]
        )

        self.ous = self.__createArrayOfObject(response, ADOU)

        return self.ous
    
    def getAllContainers(self) -> List[ADContainer]:
        if len(self.containers):
            return self.containers
        self.logger.print("Retrive all containers")

        response = self.search(
            self.defaultNamingContext,
            "(objectClass=container)",
            ldap3.SUBTREE,
            ["DistinguishedName", "name", "nTSecurityDescriptor"]
        )

        self.containers.extend(self.__createArrayOfObject(response, ADContainer))

        response = self.search(
            self.configurationNamingContext,
            "(objectClass=container)",
            ldap3.SUBTREE,
            ["DistinguishedName", "name", "nTSecurityDescriptor"]
        )

        self.containers.extend(self.__createArrayOfObject(response, ADContainer))

        return self.containers

    def getAdminSDHolder(self) -> List[ADAdminSDHolder]:
        if len(self.adminSDHolder):
            return self.adminSDHolder
        self.logger.print("Retrive adminSDHolder")

        response = self.search(
            f"CN=AdminSDHolder,CN=System,{self.defaultNamingContext}",
            "(cn=AdminSDHolder)",
            ldap3.BASE,
            ["DistinguishedName", "name", "ntSecurityDescriptor"]
        )

        self.adminSDHolder = self.__createArrayOfObject(response, ADAdminSDHolder)

        return self.adminSDHolder

    def getSchema(self) -> List[ADSchema]:
        if len(self.schema):
            return self.schema
        self.logger.print("Retrive Schema")

        # Subtree in case it's not inherant
        response = self.search(
            f"CN=Schema,{self.configurationNamingContext}",
            "(objectClass=*)",
            ldap3.SUBTREE,
            ["DistinguishedName", "name", "ntSecurityDescriptor"]
        )

        self.schema = self.__createArrayOfObject(response, ADSchema)

        return self.schema

    def getAllgMSAs(self) -> List[ADgMSA]:
        if len(self.gMSA):
            return self.gMSA
        self.logger.print("Retrive all gMSAs")
        
        response = self.search(
            self.defaultNamingContext,
            "(objectClass=msDS-GroupManagedServiceAccount)",
            ldap3.SUBTREE,
            ["DistinguishedName", "sAMAccountName", "objectSid", "ntSecurityDescriptor", "memberOf"]
        )

        self.gMSA = self.__createArrayOfObject(response, ADgMSA)

        return self.gMSA

    def getDomain(self) -> ADDomain:
        if self.domain:
            return self.domain
        self.logger.print("Retrive domain")

        response = self.search(
            self.defaultNamingContext,
            "(objectClass=*)",
            ldap3.BASE,
            ["DistinguishedName", "objectSid", "ntSecurityDescriptor"]
        )

        self.domain: ADDomain = self.__createArrayOfObject(response, ADDomain)[0]

        return self.domain

    def gatherAllInformations(self) -> None:
        self.getAllUsers()
        self.getAllGroups()
        self.getAllComputers()
        self.getAllCertificatesTemplates()
        self.getAllGPOs()
        self.getAllOUs()
        self.getAllgMSAs()
        self.getAdminSDHolder()
        self.getSchema()
        self.getDomain()

    def isPrincipalExists(self, principal: str) -> ADObject | None:
        if principal.startswith("S-1-"):
            principalObject = ADGroup.getGroupFromSid(self.groups, principal) or ADUser.getUserFromSid(self.users, principal) or ADComputer.getComputerFromSid(self.computers, principal) or ADgMSA.getgMSAFromSid(self.gMSA, principal)
        else:
            principalObject = ADGroup.getGroupSid(self.groups, principal) or ADUser.getUserSid(self.users, principal) or ADComputer.getComputerSid(self.computers, principal) or ADgMSA.getgMSASid(self.gMSA, principal)

        return principalObject

    def getImplicitGroups(self, principalObject: ADObject) -> List[ADGroup]:
         # Detect computers or users
        if isinstance(principalObject, (ADUser, ADgMSA, )):
            # Users: Domains User, Everyone, Authenticated Users
            return [
                ADGroup(
                    [WELL_KNOWN_SID.DOMAIN_USERS.name.encode()],
                    [WELL_KNOWN_SID.DOMAIN_USERS.name.encode()],
                    [WELL_KNOWN_SID.DOMAIN_USERS.name.encode()],
                    "".join([self.domain.objectSid, WELL_KNOWN_SID.DOMAIN_USERS.value]),
                    None,
                    []
                ),
                ADGroup(
                    [WELL_KNOWN_SID.EVERYONE.name.encode()],
                    [WELL_KNOWN_SID.EVERYONE.name.encode()],
                    [WELL_KNOWN_SID.EVERYONE.name.encode()],
                    WELL_KNOWN_SID.EVERYONE.value,
                    None,
                    []
                ),
                ADGroup(
                    [WELL_KNOWN_SID.AUTHENTICATED_USER.name.encode()],
                    [WELL_KNOWN_SID.AUTHENTICATED_USER.name.encode()],
                    [WELL_KNOWN_SID.AUTHENTICATED_USER.name.encode()],
                    WELL_KNOWN_SID.AUTHENTICATED_USER.value,
                    None,
                    []
                )
            ]
        elif isinstance(principalObject, ADComputer):
            # Computers: Domains Computers, Everyone, Authenticated Users
            return [
                ADGroup(
                    [WELL_KNOWN_SID.DOMAIN_COMPUTERS.name.encode()],
                    [WELL_KNOWN_SID.DOMAIN_COMPUTERS.name.encode()],
                    [WELL_KNOWN_SID.DOMAIN_COMPUTERS.name.encode()],
                    "".join([self.domain.objectSid, WELL_KNOWN_SID.DOMAIN_COMPUTERS.value]),
                    None,
                    []
                ),
                ADGroup(
                    [WELL_KNOWN_SID.EVERYONE.name.encode()],
                    [WELL_KNOWN_SID.EVERYONE.name.encode()],
                    [WELL_KNOWN_SID.EVERYONE.name.encode()],
                    WELL_KNOWN_SID.EVERYONE.value,
                    None,
                    []
                ),
                ADGroup(
                    [WELL_KNOWN_SID.AUTHENTICATED_USER.name.encode()],
                    [WELL_KNOWN_SID.AUTHENTICATED_USER.name.encode()],
                    [WELL_KNOWN_SID.AUTHENTICATED_USER.name.encode()],
                    WELL_KNOWN_SID.AUTHENTICATED_USER.value,
                    None,
                    []
                )
            ]

    def getMemberOfForPrincipal(self, principalObject: ADObject) -> List[ADObject]:        
        return principalObject.memberOf

    def getGroupRecursive(self, group: str) -> List[ADGroup]:
        groups: List[ADGroup] = list()

        groupObject: ADGroup = ADGroup.getGroupSidFromDistinguishedName(self.groups, group)

        if groupObject is None:
            self.logger.debug(f"Can't find group {group}")
            return groups
        
        groups.append(groupObject)

        if len(groupObject.memberOf):
            for group in groupObject.memberOf:
                groups.extend(self.getGroupRecursive(group))

        return groups

    def getGroupsOfPrincipal(self, principalObject: ADObject) -> set[ADGroup]:
        groups = set()

        memberOf = self.getMemberOfForPrincipal(principalObject)
        implicitMemberOf = self.getImplicitGroups(principalObject)

        # If the user have no group
        if not len(memberOf):
            return implicitMemberOf

        for group in memberOf:
            groups.update(self.getGroupRecursive(group))

        if len(implicitMemberOf):
            groups.update(implicitMemberOf)

        return groups




from impacket.krb5.kerberosv5 import getKerberosTGT, getKerberosTGS
from impacket.krb5.asn1 import AP_REQ, Authenticator, TGS_REP, seq_set
from impacket.krb5.types import Principal, KerberosTime, Ticket
from impacket.krb5.ccache import CCache
from impacket.krb5 import constants
from impacket.spnego import SPNEGO_NegTokenInit, TypesMech
from pyasn1.codec.ber import encoder, decoder
from pyasn1.type.univ import noValue

import datetime


class Kerberos:

    @staticmethod
    def kerberosLogin(target: str, user: str, password: str, domain: str = "", hashes: str = "", aesKey: str = "",
                           kdcHost: str = None, TGT=None, TGS=None, useCache: bool = False):

        if len(hashes):
            lmhash, nthash = hashes.split(":")

            if len(lmhash) % 2:
                lmhash = "0" + lmhash
            if len(nthash) % 2:
                nthash = "0" + nthash

            lmhash = bytes.fromhex(lmhash)
            nthash = bytes.fromhex(nthash)
        else:
            lmhash, nthash = "", ""

        if TGT is None or TGS is None:
            useCache = True

        targetName = "ldap/%s" % target
        if useCache:
            domain, user, TGT, TGS = CCache.parseFile(domain, user, targetName)

        # First of all, we need to get a TGT for the user
        userName = Principal(user, type=constants.PrincipalNameType.NT_PRINCIPAL.value)

        if TGT is None and TGS is None:
            tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, password, domain, lmhash, nthash,
                                                                    aesKey, kdcHost)
        else:
            tgt = TGT['KDC_REP']
            cipher = TGT['cipher']
            sessionKey = TGT['sessionKey']

        if TGS is None:
            serverName = Principal(targetName, type=constants.PrincipalNameType.NT_SRV_INST.value)
            tgs, cipher, oldSessionKey, sessionKey = getKerberosTGS(serverName, domain, kdcHost, tgt, cipher,
                                                                    sessionKey)
        else:
            tgs = TGS['KDC_REP']
            cipher = TGS['cipher']
            sessionKey = TGS['sessionKey']

        # Let's build a NegTokenInit with a Kerberos REQ_AP

        blob = SPNEGO_NegTokenInit()

        # Kerberos
        blob['MechTypes'] = [TypesMech['MS KRB5 - Microsoft Kerberos 5']]

        # Let's extract the ticket from the TGS
        tgs = decoder.decode(tgs, asn1Spec=TGS_REP())[0]
        ticket = Ticket()
        ticket.from_asn1(tgs['ticket'])

        # Now let's build the AP_REQ
        apReq = AP_REQ()
        apReq['pvno'] = 5
        apReq['msg-type'] = int(constants.ApplicationTagNumbers.AP_REQ.value)

        opts = []
        apReq['ap-options'] = constants.encodeFlags(opts)
        seq_set(apReq, 'ticket', ticket.to_asn1)

        authenticator = Authenticator()
        authenticator['authenticator-vno'] = 5
        authenticator['crealm'] = domain
        seq_set(authenticator, 'cname', userName.components_to_asn1)
        now = datetime.datetime.utcnow()

        authenticator['cusec'] = now.microsecond
        authenticator['ctime'] = KerberosTime.to_asn1(now)

        encodedAuthenticator = encoder.encode(authenticator)

        # Key Usage 11
        # AP-REQ Authenticator (includes application authenticator
        # subkey), encrypted with the application session key
        # (Section 5.5.1)
        encryptedEncodedAuthenticator = cipher.encrypt(sessionKey, 11, encodedAuthenticator, None)

        apReq['authenticator'] = noValue
        apReq['authenticator']['etype'] = cipher.enctype
        apReq['authenticator']['cipher'] = encryptedEncodedAuthenticator

        blob['MechToken'] = encoder.encode(apReq)

        return blob




from abuseACL.structures.ADObject import ADObject

class ADSchema(ADObject):

    def __init__(self, distinguishedName: str, name: str, nTSecurityDescriptor: bytes) -> None:
        self.distinguishedName      = distinguishedName[0].decode()
        self.name                   = name[0].decode()
        self.nTSecurityDescriptor   = self.parseSecurityDescriptor(nTSecurityDescriptor[0])

        self.sAMAccountName = self.name
        self.objectSid = str()




from enum import Enum


class ACCESS_MASK(Enum):
    FULL_CONTROL            = 0x000f01ff

    # Perform "Validated writes" (i.e. edit an attribute's value and have that value verified and validate by AD). The "Validated writes" is referenced by an "ObjectType GUID".
    SELF                    = 0x00000008

    # Edit one of the object's attributes. The attribute is referenced by an "ObjectType GUID".
    WRITE_PROPERTIES        = 0x00000020

    # Peform "Extended rights". "AllExtendedRights" refers to that permission being unrestricted. This right can be restricted by specifying the extended right in the "ObjectType GUID".
    ALL_EXTENDED_RIGHTS     = 0x00000100

    # Edit the object's DACL (i.e. "inbound" permissions).
    WRITE_DACL              = 0x00040000

    # Assume the ownership of the object (i.e. new owner of the victim = attacker, cannot be set to another user). 
    # With the "SeRestorePrivilege" right it is possible to specify an arbitrary owner.
    WRITE_OWNER             = 0x00080000

    # Combination of almost all other rights.
    GENERIC_ALL             = 0x10000000

    # Combination of write permissions (Self, WriteProperty) among other things.
    GENERIC_WRITE           = 0x40000000


class RIGHTS_GUID(Enum):
    # Edit the "member" attribute of the object.
    SELF_MEMBERSHIP                 = "bf9679c0-0de6-11d0-a285-00aa003049e2"

    # Change the password of the object without having to know the previous one.
    RESET_PASSWORD                  = "00299570-246d-11d0-a768-00aa006e0529"
    
    # One of the two extended rights needed to operate a DCSync.
    DS_REPLICATION_GET_CHANGES      = "1131f6aa-9c07-11d1-f79f-00c04fc2dcd2"
    
    # One of the two extended rights needed to operate a DCSync.
    DS_REPLICATION_GET_CHANGES_ALL  = "1131f6ad-9c07-11d1-f79f-00c04fc2dcd2"

    # Can be used to do Shadow Credentials
    DS_KEY_CREDENTIAL_LINK          = "5b47d60f-6090-40b2-9f37-2a4de88f3063"

    # Edit the "servicePrincipalName" attribute of the object.
    VALIDATED_SPN                   = "f3a64788-5306-11d1-a9c5-0000f80367c1"

    # All
    ALL                             = "00000000-0000-0000-0000-000000000000"

    # Edit the "gpLink" attribute of the object.
    MANAGE_GP_LINK                  = "f30e3bbf-9ff0-11d1-b603-0000f80367c1"

class WELL_KNOWN_SID(Enum):
    EVERYONE            = "S-1-1-0"
    AUTHENTICATED_USER  = "S-1-5-11"
    DOMAIN_USERS        = "-513"
    DOMAIN_COMPUTERS    = "-515"




from impacket.ldap.ldaptypes import ACCESS_ALLOWED_ACE, ACCESS_ALLOWED_OBJECT_ACE
from impacket.uuid import bin_to_string

from abuseACL.structures.structures import ACCESS_MASK, RIGHTS_GUID
from abuseACL.structures.ADObject.ADComputer import ADComputer
from abuseACL.structures.ADObject.ADGroup import ADGroup
from abuseACL.structures.ADObject.ADUser import ADUser
from abuseACL.structures.ADObject.ADgMSA import ADgMSA
from abuseACL.structures.ADObject import ADObject
from abuseACL.network.LDAP import LDAP
from abuseACL.core.Logger import Logger

class abuseACL:

    def __init__(self, ldap: LDAP, logger: Logger, extends: bool) -> None:
        self.ldap                   = ldap
        self.logger                 = logger
        self.extends                = extends

        self.users                  = self.ldap.getAllUsers()
        self.groups                 = self.ldap.getAllGroups()
        self.computers              = self.ldap.getAllComputers()
        self.certificatesTemplates  = self.ldap.getAllCertificatesTemplates()
        self.gpos                   = self.ldap.getAllGPOs()
        self.ous                    = self.ldap.getAllOUs()
        self.gMSAs                  = self.ldap.getAllgMSAs()

        self.allObjects = self.users + self.groups + self.computers + \
            self.certificatesTemplates + self.gpos + self.ous + self.gMSAs
        
        if self.extends:
            self.adminSDHolder      = self.ldap.getAdminSDHolder()
            self.schema             = self.ldap.getSchema()
            self.domain             = self.ldap.getDomain()
            self.containers         = self.ldap.getAllContainers()

            self.allObjects += self.adminSDHolder + self.schema + [self.domain] + self.containers

    def isObjectTypeGUIDRestricted(self, ace) -> RIGHTS_GUID:
        isDangerous = self.isObjectTypeGUIDDangerous(ace)

        # Restricted but has a dangerous objectTypeGUID
        if isDangerous:
            return isDangerous
        else:
            # Not restricted
            if "ObjectTypeLen" not in ace["Ace"].fields:
                return RIGHTS_GUID.ALL

        return False

    def isObjectTypeGUIDDangerous(self, ace) -> RIGHTS_GUID:
        # Check if the field exists
        if "ObjectTypeLen" in ace["Ace"].fields:
            # Check the length
            if ace["Ace"]["ObjectTypeLen"]:
                objectTypeGUID = bin_to_string(ace["Ace"]["ObjectType"]).lower()

                # Check if the right is dangerous
                for right in RIGHTS_GUID:
                    if right.value == objectTypeGUID:
                        return right

        return False

    def printVuln(self, entry: ADObject, ace, perm: ACCESS_MASK, principal: ADUser|ADComputer|ADgMSA|ADGroup, right: RIGHTS_GUID) -> None:
        self.logger.vuln(f"Result for {entry.sAMAccountName} ({entry.distinguishedName})")
        self.logger.vuln(f"    ACE Type           : {ace['Ace'].__class__.__name__}")
        self.logger.vuln(f"    Access mask        : {perm.name}")
        self.logger.vuln(f"    Principal (SID)    : {principal.sAMAccountName} ({principal.objectSid})")

        if right:
            # Right and GUID
            self.logger.vuln(f"    Object type (GUID) : {right.name} ({right.value})")

    def abuse(self, principal: ADUser|ADComputer|ADgMSA|ADGroup) -> None:
        """
        crossDomain possible check with another forest.
        Check if the user is the owner of another user, group, computer, certificateTemplate, gpo
        Check if the user have dangerous write on another user, (group, Self the user can add itself to the group), computer, certificateTemplate, gpo
        """
        haveVulnerability   = False
        principalSid        = principal.objectSid

        self.logger.debug(f"SID of the principal: {principalSid}")

        for entry in self.allObjects:
            entry: ADObject
            # Name of the Object without AD
            objectName = entry.__class__.__name__[2:]

            securityDescriptor = entry.nTSecurityDescriptor

            # Check securityDescriptor, sometimes, we can't read them like for (Microsoft Exchange container, idk why)
            if securityDescriptor is None:
                continue

            if principalSid == securityDescriptor["OwnerSid"].formatCanonical():
                haveVulnerability = True
                self.logger.vuln(f"{principal.sAMAccountName} is the owner of {entry.sAMAccountName}")

            # ACE in ACL
            for ace in securityDescriptor["Dacl"].aces:

                # Only check ALLOW
                if ace["Ace"].ACE_TYPE in [ACCESS_ALLOWED_ACE.ACE_TYPE, ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE]:
                    # Only check concerned user
                    if principalSid != ace["Ace"]["Sid"].formatCanonical():
                        continue

                    # Don't need to check if already full rights
                    if ace["Ace"]["Mask"].hasPriv(ACCESS_MASK.FULL_CONTROL.value):
                        haveVulnerability = True
                        self.printVuln(entry, ace, ACCESS_MASK.FULL_CONTROL, principal, False)
                        continue

                    for perm in ACCESS_MASK:
                        # Check if permission in current ACE
                        if not (ace["Ace"]["Mask"].hasPriv(perm.value)):
                            continue

                        vuln = f"{principal.sAMAccountName} can do {perm.name} on {entry.sAMAccountName}"
                        right = False

                        # Edit one of the object's attributes. The attribute is referenced by an "ObjectType GUID".
                        if perm.name == ACCESS_MASK.WRITE_PROPERTIES.name:
                            right = self.isObjectTypeGUIDRestricted(ace)
                            if right:
                                vuln = f"{principal.sAMAccountName} can do {perm.name}:{right} on {entry.sAMAccountName}"
                            else:
                                # Debug, in case it is useful, (No vulnerability)
                                self.logger.debug(vuln)
                                vuln = ""

                        # Perform "Extended rights". "AllExtendedRights" refers to that permission being unrestricted. This right can be restricted by specifying the extended right in the "ObjectType GUID".
                        elif perm.name == ACCESS_MASK.ALL_EXTENDED_RIGHTS.name:
                            right = self.isObjectTypeGUIDRestricted(ace)
                            if right:
                                vuln = f"{principal.sAMAccountName} can do {perm.name}:{right} on {entry.sAMAccountName}"
                            else:
                                # Debug, in case it is useful, (No vulnerability)
                                self.logger.debug(vuln)
                                vuln = ""

                        if len(vuln):
                            haveVulnerability = True

                            self.printVuln(entry, ace, perm, principal, right)

        # In case no vulnerability were found for the principal
        if not haveVulnerability:
            self.logger.error(f"Nothing found for principal {principal.sAMAccountName}")




from os import path

import setuptools

import abuseACL

cwd = path.abspath(path.dirname(__file__))

with open(path.join(cwd, "README.md")) as f:
    long_description = f.read()

with open(path.join(cwd, "requirements.txt")) as f:
    requirements = f.read()


setuptools.setup(
    name="abuseACL",
    version=abuseACL.__version__,
    description="List vulnerable ACL.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AetherBlack/abuseACL",
    author="Aether",
    classifiers=[
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 3"
    ],
    keywords="abuseACL ActiveDirectory AD",
    packages=setuptools.find_packages(),
    python_requires=">=3.6, <4",
    install_requires=requirements,
    entry_points={
        "console_scripts": ["abuseACL=abuseACL.__main__:main"]
    }
)




from impacket.examples import utils
from getpass import getpass

import argparse
import sys
import os

from abuseACL.structures.Credentials import Credentials
from abuseACL.structures.Target import Target
from abuseACL.core.abuseACL import abuseACL
from abuseACL.core.Logger import Logger
from abuseACL.network.LDAP import LDAP
from abuseACL import __banner__

class Arguments:

    debug: bool
    ts: bool
    no_pass: bool
    hashes: str
    doKerberos: bool
    aesKey: bool
    dc_ip: str
    port: int
    domain: str
    username: str
    password: str
    remote_name: str
    principal: str
    principal_sid: str
    principalsfile: str
    forest: str
    extends: bool
    doSimpleBind: bool
    recurse: bool

    def __init__(self) -> None:
        self.__parser = argparse.ArgumentParser(add_help=True, description="Automatic Windows vulnerable ACEs/ACLs listing")
        self.__parser.add_argument("-debug", default=False, action="store_true", help="Turn DEBUG output ON. (Default: False)")
        self.__parser.add_argument("-ts", action="store_true", help="Adds timestamp to every logging output")

        # Credentials
        credentials = self.__parser.add_argument_group("Credentials")
        credentials.add_argument("-no-pass", action="store_true", help="Don't ask for password (useful for -k or when using proxychains)")
        credentials.add_argument("-hashes", action="store", metavar="[LMHASH]:NTHASH", help="NT/LM hashes. LM hash can be empty.")
        credentials.add_argument("-k", action="store_true", help="Use Kerberos authentication. Grabs credentials from ccache file (KRB5CCNAME) based on target parameters. If valid credentials cannot be found, it will use the ones specified in the command line")
        credentials.add_argument("-aesKey", action="store", metavar="hex key", help="AES key to use for Kerberos Authentication (128 or 256 bits)")

        # Connection
        connection = self.__parser.add_argument_group("Connection")
        connection.add_argument("-dc-ip", action="store", metavar="ip address", help="IP Address of the domain controller. If omitted it will use the domain part (FQDN) specified in the target parameter")
        connection.add_argument("-port", type=int, action="store", help="Port of the domain controller. If omitted it will try to authenticate in LDAPS and then in LDAP.")
        connection.add_argument("-simple-bind", action="store_true", help="LDAP authentication in plaintext (SIMPLE). Sometimes, you only need to provide the domaine like CONTOSO/user instead of CONTOSO.LAN/user.")

        # User
        filters = self.__parser.add_argument_group("Filters")
        filters.add_argument("-principal", action="store", help="Find vulnerable ACE for a specific User/Computer/Group")
        filters.add_argument("-principal-sid", action="store", help="Find vulnerable ACE for a specific SID")
        filters.add_argument("-principalsfile", action="store", help="File with multiple User/Computer/Group")
        filters.add_argument("-forest", action="store", help="Forest to use if different from dc. Not implemented yet.")
        filters.add_argument("-recurse", action="store_true", help="Check group(s) related to the principal")

        # LDAP
        ldap = self.__parser.add_argument_group("LDAP")
        ldap.add_argument("-extends", action="store_true", help="Check adminSDHolder and Schema.")

        self.__parser.add_argument("target", action="store", help="[[domain/]username[:password]@]<targetName or address>")

    def parseArgs(self) -> None:
        if len(sys.argv) == 1:
            self.__parser.print_help()
            sys.exit(1)

        self._args          = self.__parser.parse_args()
        self.debug          = self._args.debug
        self.ts             = self._args.ts
        self.no_pass        = self._args.no_pass
        self.hashes         = self._args.hashes
        self.doKerberos     = self._args.k
        self.aesKey         = self._args.aesKey
        self.dc_ip          = self._args.dc_ip
        self.port           = self._args.port
        self.principal      = self._args.principal
        self.principal_sid  = self._args.principal_sid
        self.principalsfile = self._args.principalsfile
        self.forest         = self._args.forest
        self.extends        = self._args.extends
        self.doSimpleBind   = self._args.simple_bind
        self.recurse        = self._args.recurse

        self.domain, self.username, self.password, self.remote_name = utils.parse_target(self._args.target)
        if not len(self.domain):
            self.domain, self.username, self.password = utils.parse_credentials(self._args.target)
            self.remote_name = None
        
        if not len(self.password) and self.hashes is None and not self.no_pass and self.aesKey is None:
            self.password = getpass("Password:")

        if self.hashes is None:
            self.hashes = ""
        
        if ":" not in self.hashes and len(self.hashes):
            self.hashes = "aad3b435b51404eeaad3b435b51404ee:%s" % (self.hashes)
        elif len(self.hashes):
            lm, nt = self.hashes.split(":", 1)
            if not len(lm):
                self.hashes = "aad3b435b51404eeaad3b435b51404ee%s" % (self.hashes)

        if self.principalsfile:
            if not os.path.exists(self.principalsfile):
                Logger(self.debug, self.ts).error(f"File {self.principalsfile} doesn't exists")
                exit(1)

            if not os.path.isfile(self.principalsfile):
                Logger(self.debug, self.ts).error(f"File {self.principalsfile} is not a file")
                exit(1)
            
            with open(self.principalsfile) as f:
                self.principalsfile = f.read().splitlines()

def main():
    print(__banner__)

    arguments = Arguments()
    arguments.parseArgs()

    logger = Logger(arguments.debug, arguments.ts)
    credentials = Credentials(arguments.username, arguments.password, arguments.domain, arguments.hashes, arguments.aesKey, arguments.doKerberos, arguments.doSimpleBind)
    target = Target(arguments.dc_ip or arguments.remote_name or arguments.domain, arguments.port)
    ldap = LDAP(arguments.forest, target, credentials, logger)

    acl = abuseACL(ldap, logger, arguments.extends)

    logger.print("Gather all LDAP informations")
    ldap.gatherAllInformations()

    principals = arguments.principal or arguments.principal_sid or arguments.principalsfile or arguments.username

    if not isinstance(principals, list):
        principals = [principals]

    for principal in principals:
        principalObject = ldap.isPrincipalExists(principal)
        if not principalObject:
            logger.error(f"Can't find principal {principal}")
            continue

        if arguments.recurse:
            groupsOfPrincipal = ldap.getGroupsOfPrincipal(principalObject)
            logger.print(f"Principal {principalObject.sAMAccountName} is member of:")
            for group in groupsOfPrincipal:
                logger.print(f"\t{group.name} ({group.objectSid})")

            for groupObject in groupsOfPrincipal:
                acl.abuse(groupObject)

        acl.abuse(principalObject)
